/*
 * tek-getwave: get waveforms from Tektronix MDO series oscilloscopes
 * Copyright © 2020 D Scott Guthridge <scott_guthridge@rompromity.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A11 PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "archdep.h"

#include <ctype.h>
#include <errno.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "channel.h"

/*
 * channel_name_entry_t: entry in channel_name_table
 */
typedef struct channel_name_entry {
    const char	       *cn_prefix;
    const char	       *cn_suffix;
    channel_name_t	cn_channel;
} channel_name_entry_t;


/*
 * channel_encoding_table: map channel_encoding_t to encoding and width
 */
channel_encoding_entry_t channel_encoding_table[] = {
    [CH_RIB1] = { "RIBinary", 1 },
    [CH_RIB2] = { "RIBinary", 2 },
    [CH_RIB4] = { "RIBinary", 4 },
    [CH_RIB8] = { "RIBinary", 8 },
    [CH_FPB4] = { "FPbinary", 4 }
};

/*
 * channel_parameter_table: map channel_name_t to channel name and encoding
 */
channel_parameter_entry_t channel_parameter_table[] = {
    [CH_CH1]	 = { "CH1",	     CH_RIB2, CH_LF  },
    [CH_CH2]	 = { "CH2",	     CH_RIB2, CH_LF  },
    [CH_CH3]	 = { "CH3",	     CH_RIB2, CH_LF  },
    [CH_CH4]	 = { "CH4",	     CH_RIB2, CH_LF  },
    [CH_REF1]	 = { "REF1",	     CH_RIB2, CH_LF  },
    [CH_REF2]	 = { "REF2",	     CH_RIB2, CH_LF  },
    [CH_REF3]	 = { "REF3",	     CH_RIB2, CH_LF  },
    [CH_REF4]	 = { "REF4",	     CH_RIB2, CH_LF  },
    [CH_MATH]	 = { "MATH",	     CH_RIB2, CH_LF  },
    [CH_D0]	 = { "D0",	     CH_RIB1, CH_I8  },
    [CH_D1]	 = { "D1",	     CH_RIB1, CH_I8  },
    [CH_D2]	 = { "D2",	     CH_RIB1, CH_I8  },
    [CH_D3]	 = { "D3",	     CH_RIB1, CH_I8  },
    [CH_D4]	 = { "D4",	     CH_RIB1, CH_I8  },
    [CH_D5]	 = { "D5",	     CH_RIB1, CH_I8  },
    [CH_D6]	 = { "D6",	     CH_RIB1, CH_I8  },
    [CH_D7]	 = { "D7",	     CH_RIB1, CH_I8  },
    [CH_D8]	 = { "D8",	     CH_RIB1, CH_I8  },
    [CH_D9]	 = { "D9",	     CH_RIB1, CH_I8  },
    [CH_D10]	 = { "D10",	     CH_RIB1, CH_I8  },
    [CH_D11]	 = { "D11",	     CH_RIB1, CH_I8  },
    [CH_D12]	 = { "D12",	     CH_RIB1, CH_I8  },
    [CH_D13]	 = { "D13",	     CH_RIB1, CH_I8  },
    [CH_D14]	 = { "D14",	     CH_RIB1, CH_I8  },
    [CH_D15]	 = { "D15",	     CH_RIB1, CH_I8  },
    [CH_DIG32]	 = { "DIGital",	     CH_RIB4, CH_X32 },
    [CH_DIG64]	 = { "DIGital",	     CH_RIB8, CH_X64 },
    [CH_RF_AMP]	 = { "RF_AMPlitude", CH_FPB4, CH_LF  },
    [CH_RF_FREQ] = { "RF_FREQuency", CH_FPB4, CH_LF  },
    [CH_RF_PHAS] = { "RF_PHASe",     CH_FPB4, CH_LF  },
    [CH_RF_NORM] = { "RF_NORMal",    CH_FPB4, CH_LF  },
    [CH_RF_AVE]	 = { "RF_AVErage",   CH_FPB4, CH_LF  },
    [CH_RF_MAXH] = { "RF_MAXHold",   CH_FPB4, CH_LF  },
    [CH_RF_MINH] = { "RF_MINHold",   CH_FPB4, CH_LF  },
    [CH_RF_MATH] = { "MATH",	     CH_FPB4, CH_LF  },
};

/*
 * channel_name_table: map channel name to channel_name_t
 */
static channel_name_entry_t channel_name_table[] = {
    { "ch",	     "1", CH_CH1     },
    { "ch",	     "2", CH_CH2     },
    { "ch",	     "3", CH_CH3     },
    { "ch",	     "4", CH_CH4     },
    { "d",	     "0", CH_D0	     },
    { "d",	     "1", CH_D1	     },
    { "d",	     "2", CH_D2	     },
    { "d",	     "3", CH_D3	     },
    { "d",	     "4", CH_D4	     },
    { "d",	     "5", CH_D5	     },
    { "d",	     "6", CH_D6	     },
    { "d",	     "7", CH_D7	     },
    { "d",	     "8", CH_D8	     },
    { "d",	     "9", CH_D9	     },
    { "d",	    "10", CH_D10     },
    { "d",	    "11", CH_D11     },
    { "d",	    "12", CH_D12     },
    { "d",	    "13", CH_D13     },
    { "d",	    "14", CH_D14     },
    { "d",	    "15", CH_D15     },
    { "digital",      "", CH_DIG32   },
    { "digital",    "32", CH_DIG32   },
    { "digital",    "64", CH_DIG64   },
    { "math",	      "", CH_MATH    },
    { "math",	     "1", CH_MATH    },
    { "rf_math",      "", CH_RF_MATH },
    { "ref",	     "1", CH_REF1    },
    { "ref",	     "2", CH_REF2    },
    { "ref",	     "3", CH_REF3    },
    { "ref",	     "4", CH_REF4    },
    { "rf_amplitude", "", CH_RF_AMP  },
    { "rf_frequency", "", CH_RF_FREQ },
    { "rf_phase",     "", CH_RF_PHAS },
    { "rf_normal",    "", CH_RF_NORM },
    { "rf_average",   "", CH_RF_AVE  },
    { "rf_maxhold",   "", CH_RF_MAXH },
    { "rf_minhold",   "", CH_RF_MINH },
};
#define N	(sizeof(channel_name_table) / sizeof(channel_name_entry_t))

#define MAX_NAME	12

/*
 * channel_find_by_name: return the channel index by name
 *   @name: 
 */
channel_name_t
channel_find_by_name(const char *name)
{
    channel_name_t result = CH_ERROR;
    size_t length;
    char prefix[MAX_NAME + 2];
    char *suffix;
    int i;
    channel_name_entry_t *cnp;

    /*
     * Copy name into "prefix".	 If it's too long, fail.
     */
    length = strlen(name);
    if (length > MAX_NAME) {
	errno = ESRCH;
	return CH_ERROR;
    }
    (void)memcpy((void *)prefix, (void *)name, length + 1);

    /*
     * Establish suffix one byte beyond prefix.	 While prefix ends in
     * digits, move them into suffix.
     */
    *(suffix = &prefix[length + 1]) = '\000';
    while (suffix > prefix + 1 && isdigit(suffix[-2])) {
	suffix[-1] = suffix[-2];
	--suffix;
    }
    suffix[-1] = '\000';		/* re-terminate prefix */
    length = suffix - prefix - 1;	/* adjust prefix length */

    /*
     * Lower case prefix.
     */
    for (i = 0; i < length; ++i) {
	if (isupper(prefix[i])) {
	    prefix[i] = tolower(prefix[i]);
	}
    }

    /*
     * Search channel_name_table for an entry where prefix is a subset
     * of the table prefix and suffix is an exact match.  As a special
     * case, the empty prefix only matches "ch".
     */
    for (cnp = channel_name_table; cnp < &channel_name_table[N]; ++cnp) {
	if (strncmp(prefix, cnp->cn_prefix, length) == 0 &&
	    strcmp(suffix, cnp->cn_suffix) == 0 &&
	    (prefix[0] != '\000' || strcmp(cnp->cn_prefix, "ch") == 0)) {
	    if (result != CH_ERROR) {
		errno = EINVAL;
		return CH_ERROR;
	    }
	    result = cnp->cn_channel;
	}
    }
    if (result == CH_ERROR) {
	errno = ESRCH;
    }
    return result;
}
