/*
 * tek-getwave: get waveforms from Tektronix MDO series oscilloscopes
 * Copyright © 2020 D Scott Guthridge <scott_guthridge@rompromity.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A11 PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef CHANNEL_H
#define CHANNEL_H

#include <stdint.h>

/*
 * channel_name_t: instrument output channel
 */
typedef enum {
    CH_ERROR = -1,
    CH_CH1, CH_CH2, CH_CH3, CH_CH4,
    CH_REF1, CH_REF2, CH_REF3, CH_REF4,
    CH_MATH,
    CH_D0, CH_D1, CH_D2, CH_D3, CH_D4, CH_D5, CH_D6, CH_D7,
    CH_D8, CH_D9, CH_D10, CH_D11, CH_D12, CH_D13, CH_D14, CH_D15,
    CH_DIG32, CH_DIG64,
    CH_RF_AMP, CH_RF_FREQ, CH_RF_PHAS,
    CH_RF_NORM, CH_RF_AVE, CH_RF_MAXH, CH_RF_MINH,
    CH_RF_MATH
} channel_name_t;

/*
 * channel_value_t: type and format of resulting channel data
 */
typedef enum {
    CH_I8,		/* int8_t, decimal */
    CH_X32,		/* uint32_t, hexadecimal */
    CH_X64,		/* uint64_t, hexadecimal */
    CH_LF		/* double */
} channel_format_t;

/*
 * channel_data: instrument curve data
 */
typedef struct channel_data {
    channel_name_t	cd_channel;
    uint32_t		cd_rows;
    uint32_t		cd_columns;
    char	       *cd_description;
    char	       *cd_xunit;
    char	       *cd_yunit;
    int			cd_pt_off;
    double		cd_xincr;
    double		cd_xzero;
    channel_format_t	cd_format;
    union {
	int8_t   *s8;
	uint32_t *u32;
	uint64_t *u64;
	double   *lf;
    }			cd_data;
} channel_data_t;

/*
 * channel_encoding_t: encoding of data returned from instrument
 */
typedef enum {
    CH_RIB1,	/* signed integer  8 */
    CH_RIB2,	/* signed integer 16 */
    CH_RIB4,	/* signed integer 32 */
    CH_RIB8,	/* signed integer 64 */
    CH_FPB4	/* IEEE float 32 */
} channel_encoding_t;

/*
 * channel_encoding_entry_t: entry in channel_encoding_table
 */
typedef struct channel_encoding_entry {
    const char	       *ce_encoding;
    int8_t		ce_width;
} channel_encoding_entry_t;

/*
 * channel_parameter_entry_t: entry in channel_parameter_table
 */
typedef struct channel_parameter {
    const char	       *ch_channel_name;
    channel_encoding_t	ch_encoding;
    channel_format_t	ch_format;
} channel_parameter_entry_t;

extern channel_encoding_entry_t channel_encoding_table[];
extern channel_parameter_entry_t channel_parameter_table[];
extern channel_name_t channel_find_by_name(const char *name);

#endif /* CHANNEL_H */
