/*
 * dstring: maintain dynamically growing buffers of text or binary data
 * Copyright (C) 2000,2020 D Scott Guthridge <scott_guthridge@rompromity.net>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "archdep.h"

#include <assert.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include "dstring.h"

#ifndef MIN
#define MIN(x, y)	((x) <= (y) ? (x) : (y))
#endif /* MIN */

#ifndef MAX
#define MAX(x, y)	((x) >= (y) ? (x) : (y))
#endif /* MAX */

/*
 * ds_init: initialize a dstring object and allocate the initial buffer
 *   @dsp: dstring object to initialize
 *   @initital_allocation: size of the initial allocation
 */
int
ds_init(dstring_type *dsp, size_t initial_allocation)
{
    (void)memset((void *)dsp, 0, sizeof(dstring_type));
    if (initial_allocation != 0) {
	if ((dsp->ds_base = malloc(initial_allocation)) == NULL) {
	    return -1;
	}
	dsp->ds_allocation = initial_allocation;
    }
    return 0;
}

/*
 * ds_free: free any memory held by a dstring object
 *   @dsp: dstring object
 */
void
ds_free(dstring_type *dsp)
{
    free((void *)dsp->ds_base);
    dsp->ds_base = NULL;
    dsp->ds_allocation = 0;
    dsp->ds_length = 0;
    dsp->ds_error = EINVAL;
}

/*
 * ds_extend: extend the allocation as needed to hold additional_length
 *   @dsp: dstring object
 *   @additional_length: additional bytes to hold
 */
static char *
ds_extend(dstring_type *dsp, size_t additional_length)
{
    if (dsp->ds_length + additional_length + 1 > dsp->ds_allocation) {
	size_t temp = dsp->ds_length + additional_length + 1;
	size_t new_allocation = MAX(temp + (temp >> 1), 16);

	if ((dsp->ds_base = (char *)realloc((void *)dsp->ds_base,
					    new_allocation)) == NULL) {
	    dsp->ds_error = errno;
	    dsp->ds_allocation = 0;
	    dsp->ds_length = 0;
	    return NULL;
	}
	(void)memset((void *)&dsp->ds_base[dsp->ds_allocation], 0,
		     new_allocation - dsp->ds_allocation);
	dsp->ds_allocation = new_allocation;
    }
    return dsp->ds_base;
}

/*
 * ds_add_char: add a single character
 *   @dsp: dstring object
 *   @c: character to add
 */
char *
ds_add_char(dstring_type *dsp, char c)
{
    if (dsp->ds_error != 0) {
	return NULL;
    }
    if (ds_extend(dsp, 1) == NULL) {
	return NULL;
    }
    dsp->ds_base[dsp->ds_length++] = c;
    return dsp->ds_base;
}

/*
 * ds_add_binary: append bytes
 *   @dsp: dstring object
 *   @vp: data to append
 *   @length: count of bytes to append
 */
char *
ds_add_binary(dstring_type *dsp, void *vp, size_t length)
{
    if (ds_extend(dsp, length) == NULL) {
	return NULL;
    }
    (void)memcpy((void *)&dsp->ds_base[dsp->ds_length], vp, length);
    dsp->ds_length += length;
    return dsp->ds_base;
}

/*
 * ds_add_string: append a string
 *   @dsp: dstring object
 *   @cp: string to append
 */
char *
ds_add_string(dstring_type *dsp, const char *cp)
{
    return ds_add_binary(dsp, (void *)cp, strlen(cp));
}

/*
 * ds_set_length: set the user-data length
 *   @dsp: dstring object
 *   @new_length: total number of user-data bytes
 */
char *
ds_set_length(dstring_type *dsp, size_t new_length)
{
    /*
     * Setting the length to zero clears any errors.  Otherwise,
     * if there is an error pending, return failure.
     */
    if (new_length == 0) {
	dsp->ds_error = 0;
    } else if (dsp->ds_error != 0) {
	return NULL;
    }

    /*
     * If increasing the length, make sure there is sufficient allocation
     * for the additional length and a terminating NUL character.  If
     * decreasing the length, fill the vacated area with NUL chars.
     */
    if (new_length > dsp->ds_length) {
	if (ds_extend(dsp, new_length - dsp->ds_length) == NULL) {
	    return NULL;
	}
    } else if (new_length < dsp->ds_length) {
	(void)memset((void *)&dsp->ds_base[new_length], 0,
		     dsp->ds_length - new_length);
    }
    dsp->ds_length = new_length;
    return dsp->ds_base;
}

/*
 * ds_printf: sprintf into the dstring object
 *   @dsp: dstring object
 *   @format: format string
 */
char *
ds_printf(dstring_type *dsp, const char *format, ...)
{
    va_list ap;
    char *rv;

    if (dsp->ds_error != 0) {
	return NULL;
    }
    va_start(ap, format);
    rv = ds_vprintf(dsp, format, ap);
    va_end(ap);

    return rv;
}

/*
 * ds_printf: vsprintf into the dstring object
 *   @dsp: dstring object
 *   @format: format string
 *   @ap: standard argument list
 */
char *
ds_vprintf(dstring_type *dsp, const char *format, va_list ap)
{
    int rc;
    char *str = NULL;

    if (dsp->ds_error != 0) {
	return NULL;
    }
    if ((rc = vasprintf(&str, format, ap)) == -1) {
	dsp->ds_error = errno;
	return NULL;
    }
    if (ds_extend(dsp, rc) == NULL) {
	free((void *)str);
	return NULL;
    }
    (void)memcpy((void *)&dsp->ds_base[dsp->ds_length], (void *)str, rc);
    dsp->ds_length += rc;
    free((void *)str);
    return dsp->ds_base;
}

/*
 * ds_fgets: read a line from the given file (removing the newline char)
 *   @dsp: dstring object
 *   @fp: FILE pointer
 *   @max_length: fail if string exceeds max_length (not including the NUL)
 */
char *
ds_fgets(dstring_type *dsp, FILE *fp, size_t max_length)
{
    size_t remaining_length = max_length;

    if (dsp->ds_error != 0) {
	return NULL;
    }
    for (;;) {
	size_t chunk_length = MIN(remaining_length, 1023);
	size_t length;
	char *current;
	char *cp;

	/*
	 * Make enough room for the data plus newline character.
	 */
	if (ds_extend(dsp, chunk_length + 1) == NULL) {
	    return NULL;
	}

	/*
	 * Read from the file.  Add one for the newline and one
	 * for the NUL character (which overwrites the existing
	 * NUL).
	 */
	current = &dsp->ds_base[dsp->ds_length];
	if (fgets(current, chunk_length + 2, fp) == NULL) {
	    break;
	}

	/*
	 * If we got the newline char, overwrite it with a NUL and update
	 * the length.
	 */
	if ((cp = strchr(current, '\n')) != NULL) {
	    *cp = '\000';
	    dsp->ds_length   += cp - current;
	    remaining_length -= cp - current;
	    break;
	}

	/*
	 * Otherwise, update the length with the amount read and loop.  If
	 * remaining length is zero, then the string exceeds the allowed
	 * length.
	 */
	length = strlen(current);
	dsp->ds_length   += length;
	if (length >= remaining_length) {
	    dsp->ds_error = EMSGSIZE;
	    break;
	}
	remaining_length -= length;
    }
    /*
     * Return NULL on error or EOF.  Caller has to call ds_get_error
     * to disambiguate.
     */
    if (dsp->ds_error != 0 || remaining_length == max_length) {
	return NULL;
    }
    return dsp->ds_base;
}

/*
 * ds_fputs: write a line to the given file (adding a newline char)
 *   @dsp: dstring object
 *   @fp: FILE pointer
 */
int
ds_fputs(dstring_type *dsp, FILE *fp)
{
    if (dsp->ds_error != 0) {
	return -1;
    }
    if (fputs(dsp->ds_base, fp) == -1) {
	dsp->ds_error = (errno != 0) ? errno : EIO;
	return -1;
    }
    return 0;
}

/*
 * ds_fread: read a record from a file
 *   @dsp: dstring object
 *   @fd: file descriptor
 *   @size: read this many bytes (stop on short read)
 */
char *
ds_fread(dstring_type *dsp, FILE *fp, size_t size)
{
    int rc;
    char *current;

    /*
     * Make enough room for the read.
     */
    if (ds_extend(dsp, size) == NULL) {
	return NULL;
    }

    /*
     * Read.  Pass the size as number of records so that the return
     * value
     */
    current = &dsp->ds_base[dsp->ds_length];
    rc = fread((void *)current, /*size=*/1, /*nrec=*/size, fp);
    assert(rc >= 0);
    dsp->ds_length += rc;
    if (rc != size && ferror(fp)) {
	dsp->ds_error = EIO;
	return NULL;
    }
    return dsp->ds_base;
}

/*
 * ds_fread: read a record from a file
 *   @dsp: dstring object
 *   @fd: file descriptor
 *   @size: read this many bytes (stop on short read)
 */
int
ds_fwrite(dstring_type *dsp, FILE *fp)
{
    int rc;

    rc = fwrite((void *)dsp->ds_base, /*size=*/1, /*nrec=*/dsp->ds_length, fp);
    if (rc != dsp->ds_length) {
	dsp->ds_error = (errno != 0) ? errno : EIO;
	return -1;
    }
    return 0;
}

/*
 * ds_read: read length bytes from a file descriptor
 *   @dsp: dstring object
 *   @fd: file descriptor
 *   @length: read at least this many bytes (unless EOF or error)
 */
char *
ds_read(dstring_type *dsp, int fd, size_t length)
{
    if (dsp->ds_error != 0) {
	return NULL;
    }
    while (length != 0) {
	size_t chunk_length = MIN(length, 1024);
	char *current;
	int rc;

	/*
	 * Make enough room for the read.
	 */
	if (ds_extend(dsp, chunk_length) == NULL) {
	    return NULL;
	}

	/*
	 * Read
	 */
	current = &dsp->ds_base[dsp->ds_length];
	if ((rc = read(fd, current, chunk_length)) == -1) {
	    dsp->ds_error = errno;
	    return NULL;
	}
	if (rc == 0) {
	    return NULL;
	}
	dsp->ds_length += rc;
	length         -= rc;
    }
    return dsp->ds_base;
}

/*
 * ds_write: write a dstring object to a file
 *   @dsp: dstring object
 *   @fd: file descriptor
 */
int
ds_write(dstring_type *dsp, int fd)
{
    int rc;
    char *current;
    size_t remaining;

    if (dsp->ds_error != 0) {
	return -1;
    }
    current = dsp->ds_base;
    remaining = dsp->ds_length;
    while (remaining > 0) {
	if ((rc = write(fd, current, remaining)) == -1) {
	    dsp->ds_error = errno;
	    return -1;
	}
	current += rc;
	remaining -= rc;
    }
    return 0;
}

/*
 * ds_take_string: take the buffer away from the dstring object
 *   @dsp: dstring object
 *
 *   Caller must free the returned storage with a call to free.
 */
char *
ds_take_string(dstring_type *dsp)
{
    char *result;

    if (dsp->ds_error != 0) {
	return NULL;
    }
    result = dsp->ds_base;
    dsp->ds_base = NULL;
    dsp->ds_length = 0;
    dsp->ds_allocation = 0;

    return result;
}
