/*
 * dstring: maintain dynamically growing buffers of text or binary data
 * Copyright (C) 2000,2020 D Scott Guthridge <scott_guthridge@rompromity.net>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * The dstring object maintains an dynamically-expandable buffer of
 * text or binary data.  User data is stored starting at the buffer
 * base, extending for length bytes.  Following the user data are NUL
 * characters (at least one) up to allocation, so that text data are
 * always NUL-terminated.
 *
 *	DDDDDDDDDDDDD0000000000000
 *	|            |            |
 *	base         length       allocation
 *
 * The buffer is reallocated as needed as data is added.  Avoid writing
 * data beyond the current length as it will not be preserved when the
 * buffer is resized.
 *
 * If an error is encountered, subsequent modifications to the buffer are
 * ignored until the error is cleared, either by a call to ds_clear_error
 * or a call to ds_set_length with a length of zero.  Because errors are
 * "sticky", you can perform a series of modifications to the string
 * ignoring errors, then check once at the end if an error has occurred.
 *
 * Most of the dstring functions return the current buffer base on success
 * or NULL on error.
 */

#ifndef DSTRING_H
#define DSTRING_H

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>

/*
 * dstring_type: dstring object -- treat as opaque
 */
typedef struct dstring {
    char *ds_base;
    size_t ds_length;
    size_t ds_allocation;
    int ds_error;
} dstring_type;

extern int ds_init(dstring_type *dsp, size_t initial_allocation);
extern void ds_free(dstring_type *dsp);
extern char *ds_add_char(dstring_type *dsp, char c);
extern char *ds_add_binary(dstring_type *dsp, void *dp, size_t length);
extern char *ds_add_string(dstring_type *dsp, const char *cp);
extern char *ds_set_length(dstring_type *dsp, size_t length);
extern char *ds_printf(dstring_type *dsp, const char *format, ...);
extern char *ds_vprintf(dstring_type *dsp, const char *format, va_list ap);
extern char *ds_fgets(dstring_type *dsp, FILE *fp, size_t max_length);
extern int   ds_fputs(dstring_type *dsp, FILE *fp);
extern char *ds_fread(dstring_type *dsp, FILE *fp, size_t length);
extern int   ds_fwrite(dstring_type *dsp, FILE *fp);
extern char *ds_read(dstring_type *dsp, int fd, size_t length);
extern int   ds_write(dstring_type *dsp, int fd);
extern char *ds_take_string(dstring_type *dsp);

/*
 * ds_get_base: return the current address of the buffer
 *   @dsp: dstring object
 */
static inline char *ds_get_base(dstring_type *dsp)
{
    return dsp->ds_base;
}

/*
 * ds_get_length: return the number of bytes of user-data
 *   @dsp: dstring object
 */
static inline size_t ds_get_length(dstring_type *dsp)
{
    return dsp->ds_length;
}

/*
 * ds_get_error: return an errno value if the dstring has experienced an error
 *   @dsp: dstring object
 */
static inline int ds_get_error(dstring_type *dsp)
{
    return dsp->ds_error;
}

/*
 * ds_clear: set the data length to zero
 */
static inline void ds_clear(dstring_type *dsp)
{
    ds_set_length(dsp, 0);
}

/*
 * ds_clear_error: clear the error value
 *   @dsp: dstring object
 */
static inline void ds_clear_error(dstring_type *dsp)
{
    dsp->ds_error = 0;
}

#endif /* DSTRING_H */
