/*
 * tek-getwave: get waveforms from Tektronix MDO series oscilloscopes
 * Copyright © 2020 D Scott Guthridge <scott_guthridge@rompromity.net>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A11 PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "archdep.h"

#include <arpa/inet.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <inttypes.h>
#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "channel.h"
#include "dstring.h"

char *progname;

#define MAX_LINE	1024

/*
 * options
 */
static char options[] = "h:p:";
static char const *const usage[] = {
	"[-h hostname] [-p port] channels...",
	NULL
};
static char const *const help[] = {
	"-h hostname   specify the oscilloscope",
	"-p port       port number (default 4000)",
	"channels...  {ch1|ch2|ref1|ref2|math|d0..d15|binary|rf_norm|rf_ave|rf_maxh|rf_minh|...}",
	NULL
};
static char *h_opt = "mdo3102";
static char *p_opt = "4000";

/*
 * reply_buffer: global buffer used to read replies from the instrument
 */
static dstring_type reply_buffer;

/*
 * print_usage
 */
static void
print_usage(void)
{
	char const * const *cpp;

	for (cpp = usage; *cpp != NULL; ++cpp) {
		(void)fprintf(stderr, "%s: %s\n", progname, *cpp);
	}
	for (cpp = help; *cpp != NULL; ++cpp) {
		(void)fprintf(stderr, "%s\n", *cpp);
	}
}

/*
 * read_status: query the device status and report any events
 *   @fp: instrument stream
 */
static int
read_status(FILE *fp)
{
    char *base;
    char *tail;
    int result;
    dstring_type status_buffer;

    /*
     * Use a private buffer for status replies so that we can
     * read status without destroying the contents of reply_buffer.
     */
    if (ds_init(&status_buffer, 64) != 0) {
	return -1;
    }

    /*
     * Request status and report exceptions.
     */
    (void)fprintf(fp, "*ESR?\n");
    if ((base = ds_fgets(&status_buffer, fp, MAX_LINE)) == NULL) {
	return -1;
    }
    result = (int)strtol(base, &tail, 0);
    if (tail == base || *tail != '\000') {
	errno = EPROTO;
	return -1;
    }
    if (result != 0) {
	(void)fprintf(fp, "EVQty?\n");
	ds_clear(&status_buffer);
	if ((base = ds_fgets(&status_buffer, fp, MAX_LINE)) != NULL) {
	    int evqty = atoi(base);
	    int i;

	    for (i = 0; i < evqty; ++i) {
		(void)fprintf(fp, "EVMSg?\n");
		ds_clear(&status_buffer);
		if ((base = ds_fgets(&status_buffer, fp, MAX_LINE)) == NULL) {
		    break;
		}
		(void)fprintf(stderr, "%s: %s:%s: %s\n",
			      progname, h_opt, p_opt, base);
	    }
	}
    }
    return result;
}

/*
 * read_block: read a binary block from the given file and append to dstring
 *   @fp: instrument stream
 *   @max_length: limit on the maximum block length
 *
 * Return:
 *    0: success
 *   -1: error
 */
static int
read_block(FILE *fp, uint64_t max_length)
{
    int c;
    uint64_t n, u;
    char *cp;
    char tmpbuf[17];

    /*
     * If the first character is not a '#', then we don't have a
     * binary block.  Put the first character back (if one taken)
     * and return with ENOMSG.
     */
    errno = 0;
    if ((c = getc(fp)) == EOF) {
	errno = ENOMSG;
	return -1;
    }
    if (c != '#') {
	ungetc(c, fp);
	errno = ENOMSG;
	return  -1;
    }

    /*
     * Read the length of the length character (one hex digit).
     */
    errno = 0;
    if ((c = getc(fp)) == EOF) {
	errno = EPROTO;
	return -1;
    }
    tmpbuf[0] = c;
    tmpbuf[1] = '\000';
    n = (uint64_t)strtoull(tmpbuf, &cp, 16);
    if (cp == tmpbuf) {
	errno = EPROTO;
	return -1;
    }
    assert(n < 16);

    /*
     * Read the length and check against the given limit.
     */
    for (u = 0; u < n; ++u) {
	if ((c = getc(fp)) == EOF) {
	    errno = EPROTO;
	    return -1;
	}
	tmpbuf[u] = c;
    }
    tmpbuf[u] = 0;
    n = (uint64_t)strtoull(tmpbuf, &cp, 10);
    if (n > max_length) {
	errno = EMSGSIZE;
	return -1;
    }

    /*
     * Read the binary data.
     */
    for (u = 0; u < n; ++u) {
	if ((c = getc(fp)) == EOF) {
	    errno = EPROTO;
	    return -1;
	}
	ds_add_char(&reply_buffer, c);
    }

    /*
     * Expect a newline character.
     */
    if ((c = getc(fp)) != '\n') {
	(void)fprintf(stderr, "%s: newline missing at end of block\n",
		      progname);
	if (c != EOF) {
	    ungetc(c, fp);
	}
    }
    return 0;
}

/*
 * rib8_to_int8: convert an IEEE 2's complement signed byte to int8_t
 *   @cp: address of byte
 */
int8_t
rib8_to_int8(char *cp)
{
    uint8_t u8 = *(uint8_t *)cp;

    return (int8_t)((u8 & UINT8_C(0x80)) ?
	-((~u8 + 1) & UINT8_C(0xff)) : u8);
}

/*
 * rib16_to_int16: convert a 2's complement signed 2-byte word to int16_t
 *   @cp: address of first byte
 */
int16_t
rib16_to_int16(char *cp)
{
    uint16_t u16 = ntohs(*(uint16_t *)cp);

    return (int16_t)((u16 & UINT16_C(0x8000)) ?
    	-((~u16 + 1) & UINT16_C(0xffff)) : u16);
}

/*
 * rib32_to_int32: convert a 2's complement signed 4-byte word to int32_t
 *   @cp: address of first byte
 */
int32_t
rib32_to_int32(char *cp)
{
    uint32_t u32 = ntohl(*(uint32_t *)cp);

    return (int32_t)((u32 & UINT32_C(0x80000000)) ?
    	-((~u32 + 1) & UINT32_C(0xffffffff)) : u32);
}

#ifndef HAVE_NTOHLL
/*
 * ntohll: convert big-endian u64_t to native u64_t
 *   @n64: big-endian u64_t value
 */
static uint64_t
ntohll(uint64_t n64)
{
    uint32_t msw = 0[(uint32_t *)&n64];
    uint32_t lsw = 1[(uint32_t *)&n64];

    return ((uint64_t)ntohl(msw) << 32) | (uint64_t)ntohl(lsw);
}
#endif /* HAVE_NTOHLL */

/*
 * rib64_to_int64: convert a 2's complement signed 4-byte word to int64_t
 *   @cp: address of first byte
 */
int64_t
rib64_to_int64(char *cp)
{
    uint64_t u64 = ntohll(*(uint64_t *)cp);

    return (int64_t)((u64 & UINT64_C(0x8000000000000000)) ?
    	-((~u64 + 1) & UINT64_C(0xffffffffffffffff)) : u64);
}

/*
 * fpb32_to_double: convert an IEEE754 32-byte float to native double
 *   @cp: address of first byte
 */
double
fpb32_to_double(char *cp)
{
    uint32_t u32 = ntohl(*(uint32_t *)cp);

    return (double)*(float *)&u32;
}

/*
 * read_waveform: read the indicated waveform
 */
static channel_data_t *
read_waveform(FILE *fp, channel_name_t channel)
{
    const channel_parameter_entry_t *chp = &channel_parameter_table[channel];
    const channel_encoding_entry_t *cep;
    uint32_t nr_pt = 0;
    channel_data_t *cdp;
    char *base;
    double yoff, ymult, yzero;
    size_t data_length;
    int i;

    /*
     * Allocate the channel_data_t.
     */
    if ((cdp = malloc(sizeof(channel_data_t))) == NULL) {
	return NULL;
    }
    (void)memset((void *)cdp, 0, sizeof(channel_data_t));

    cep = &channel_encoding_table[chp->ch_encoding];
    (void)fprintf(fp, "DATa INIT\n");
    (void)fprintf(fp, "DATa:SOUrc %s\n", chp->ch_channel_name);
    (void)fprintf(fp, "DATa:ENCdg %s\n", cep->ce_encoding);
    (void)fprintf(fp, "WFMOutpre:BYT_Nr %d\n", cep->ce_width);
    if (read_status(fp) != 0) {
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:NR_Pt?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    nr_pt = atoi(base);

    (void)fprintf(fp, "WFMOutpre:PT_Fmt?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    if (strcmp(base, "Y") == 0) {
	cdp->cd_columns = 1;
    } else if (strcmp(base, "ENV") == 0) {
	cdp->cd_columns = 2;
    } else {
	(void)fprintf(stderr, "%s: read_waveform: unknown PT_Fmt: %s\n",
		      progname, base);
	errno = EPROTO;
	free((void *)cdp);
	return NULL;
    }
    assert(nr_pt % cdp->cd_columns == 0);
    cdp->cd_rows = nr_pt / cdp->cd_columns;

    (void)fprintf(fp, "WFMOutpre:WFId?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    if ((cdp->cd_description = strdup(base)) == NULL) {
	errno = ENOMEM;
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:XUNit?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    if ((cdp->cd_xunit = strdup(base)) == NULL) {
	errno = ENOMEM;
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:YUNit?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    if ((cdp->cd_yunit = strdup(base)) == NULL) {
	errno = ENOMEM;
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:PT_Off?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    cdp->cd_pt_off = atoi(base);

    (void)fprintf(fp, "WFMOutpre:XINcr?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    cdp->cd_xincr = (double)cdp->cd_columns * atof(base);
    if (cdp->cd_xincr == 0.0) {
	(void)fprintf(stderr, "%s: read_waveform: invalid X_INcr: %s\n",
		      progname, base);
	errno = EPROTO;
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:XZEro?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    cdp->cd_xzero = atof(base);

    (void)fprintf(fp, "WFMOutpre:YOFf?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    yoff = atof(base);

    (void)fprintf(fp, "WFMOutpre:YMUlt?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    ymult = atof(base);
    if (ymult == 0.0) {
	(void)fprintf(stderr, "%s: read_waveform: invalid YMUlt: %s\n",
		      progname, base);
	errno = EPROTO;
	free((void *)cdp);
	return NULL;
    }

    (void)fprintf(fp, "WFMOutpre:YZEro?\n");
    ds_clear(&reply_buffer);
    if ((base = ds_fgets(&reply_buffer, fp, MAX_LINE)) == NULL) {
	free((void *)cdp);
	return NULL;
    }
    yzero = atof(base);

    cdp->cd_format = chp->ch_format;

    (void)fprintf(fp, "CURVe?\n");
    ds_clear(&reply_buffer);
    data_length = nr_pt * cep->ce_width;
    if (read_block(fp, data_length) != 0) {
	free((void *)cdp);
	return NULL;
    }
    base = ds_get_base(&reply_buffer);
    if (ds_get_length(&reply_buffer) != data_length) {
	errno = EMSGSIZE;
	free((void *)cdp);
	return NULL;
    }

    /*
     * Allocate the output vector.
     */
    switch (cdp->cd_format) {
    case CH_I8:
	if ((cdp->cd_data.s8 = malloc(nr_pt * sizeof(int8_t))) == NULL) {
	    free((void *)cdp);
	    return NULL;
	}
	break;

    case CH_X32:
	if ((cdp->cd_data.u32 = malloc(nr_pt * sizeof(uint32_t))) == NULL) {
	    free((void *)cdp);
	    return NULL;
	}
	break;

    case CH_X64:
	if ((cdp->cd_data.u64 = malloc(nr_pt * sizeof(uint64_t))) == NULL) {
	    free((void *)cdp);
	    return NULL;
	}
	break;

    case CH_LF:
	if ((cdp->cd_data.lf = malloc(nr_pt * sizeof(double))) == NULL) {
	    return NULL;
	}
	break;

    default:
	abort();
	/*NOTREACHED*/
    }


    if (strcmp(cep->ce_encoding, "RIBinary") == 0) {
	switch (cdp->cd_format) {
	case CH_I8:
	    for (i = 0; i < nr_pt; ++i) {
		cdp->cd_data.s8[i] = rib8_to_int8(&base[i]);
	    }
	    break;

	case CH_X32:
	    for (i = 0; i < nr_pt; ++i) {
		cdp->cd_data.u32[i] = rib32_to_int32(&base[4 * i]);
	    }
	    break;

	case CH_X64:
	    for (i = 0; i < nr_pt; ++i) {
		cdp->cd_data.u64[i] = rib64_to_int64(&base[8 * i]);
	    }
	    break;

	case CH_LF:
	    for (i = 0; i < nr_pt; ++i) {
		cdp->cd_data.lf[i] = yzero +
		    ymult * ((double)rib16_to_int16(&base[2 * i]) - yoff);
	    }
	    break;

	default:
	    abort();
	    /*NOTREACHED*/
	}
	return cdp;

    }
    if (strcmp(cep->ce_encoding, "FPbinary") == 0) {
	assert(cdp->cd_format == CH_LF);
	for (i = 0; i < nr_pt; ++i) {
	    cdp->cd_data.lf[i] = fpb32_to_double(&base[4 * i]);
	}
	return cdp;
    }
    abort();
    /*NOTREACHED*/
}

/*
 * free_waveform
 */
static void
free_waveform(channel_data_t *cdp)
{
    free((void *)cdp->cd_description);
    free((void *)cdp->cd_xunit);
    free((void *)cdp->cd_yunit);
    switch (cdp->cd_format) {
    case CH_I8:
	free((void *)cdp->cd_data.s8);
	break;
    case CH_X32:
	free((void *)cdp->cd_data.u32);
	break;
    case CH_X64:
	free((void *)cdp->cd_data.u64);
	break;
    case CH_LF:
	free((void *)cdp->cd_data.lf);
	break;
    }
    free((void *)cdp);
}

/*
 * main
 */
int
main(int argc, char **argv)
{
    int rc;
    int sock;
    int status;
    int i;
    struct addrinfo ai_hints;
    struct addrinfo *aip_list = NULL;
    struct addrinfo *aip;
    FILE *fp;

    /*
     * Parse options.
     */
    if ((char *)NULL == (progname = strrchr(argv[0], '/'))) {
	progname = argv[0];
    } else {
	++progname;
    }
    for (;;) {
	switch (getopt(argc, argv, options)) {
	case 'h':
	    h_opt = optarg;
	    continue;

	case 'p':
	    p_opt = optarg;
	    continue;

	case -1:
	    break;

	default:
	    print_usage();
	    exit(1);
	    /*NOTREACHED*/
	}
	break;
    }
    argc -= optind;
    argv += optind;

    /*
     * Init the global reply buffer.
     */
    if (ds_init(&reply_buffer, 100) != 0) {
	(void)fprintf(stderr, "%s: ds_init: %s\n", progname, strerror(errno));
	exit(2);
	/*NOTREACHED*/
    }

    /*
     * Connect to the scope.
     */
    (void)memset((void *)&ai_hints, 0, sizeof(ai_hints));
    ai_hints.ai_family = AF_UNSPEC;
    ai_hints.ai_socktype = SOCK_STREAM;
    ai_hints.ai_flags = 0;
    ai_hints.ai_protocol = 0;
    if ((rc = getaddrinfo(h_opt, p_opt, &ai_hints, &aip_list)) != 0) {
	(void)fprintf(stderr, "%s: getaddrinfo: %s:%s: %s\n",
	    progname, h_opt, p_opt, gai_strerror(rc));
	exit(3);
	/*NOTREACHED*/
    }
    for (aip = aip_list; aip != NULL; aip = aip->ai_next) {
	sock = socket(aip->ai_family, aip->ai_socktype, aip->ai_protocol);
	if (sock == -1)
	    continue;

	if (connect(sock, aip->ai_addr, aip->ai_addrlen) == -1) {
	    close(sock);
	    continue;
	}
	break;
    }
    if (aip == NULL) {
	(void)fprintf(stderr, "%s: connect: %s:%s: %s\n",
	    progname, h_opt, p_opt, strerror(errno));
	exit(4);
	/*NOTREACHED*/
    }
    freeaddrinfo(aip_list);
    aip_list = NULL;
    if ((fp = fdopen(sock, "r+")) == NULL) {
	(void)fprintf(stderr, "%s: fdopen: %s:%s: %s\n",
		      progname, h_opt, p_opt, strerror(errno));
	exit(5);
	/*NOTREACHED*/
    }
    setvbuf(fp, NULL, _IOLBF, 0);

    /*
     * Set instrument modes.
     */
    (void)fprintf(fp, "\n*CLS\n");
    (void)fprintf(fp, "DESE 60\n");	/* CME|EXE|DDE|QYE */
    (void)fprintf(fp, "HEADer OFF\n");
    (void)fprintf(fp, "VERBose OFF\n");
    (void)fprintf(fp, "SOCKETServer:PROTOCol NONe\n");
    if ((status = read_status(fp)) == -1) {
	(void)fprintf(stderr, "%s: read_status: %s:%s: %s\n",
	    progname, h_opt, p_opt, strerror(errno));
	exit(6);
	/*NOTREACHED*/
    }
    if (status != 0) {
	(void)fprintf(stderr, "%s: read_status: %s:%s: 0x%02x\n",
	    progname, h_opt, p_opt, status);
	exit(7);
	/*NOTREACHED*/
    }

    /*
     * Get each requested channel.
     */
    for (i = 0; i < argc; ++i) {
	channel_name_t channel;
	channel_data_t *cdp;
	int row, column;
	
	if ((channel = channel_find_by_name(argv[i])) == -1) {
	    (void)fprintf(stderr, "%s: invalid channel: %s\n",
		progname, argv[i]);
	    exit(8);
	    /*NOTREACHED*/
	}
	if ((cdp = read_waveform(fp, channel)) == NULL) {
	    (void)fprintf(stderr, "%s: read_waveform: %s\n",
	    	progname, strerror(errno));
	    exit(9);
	    /*NOTREACHED*/
	}
	(void)printf("# %s\n", cdp->cd_description);
	for (row = 0; row < cdp->cd_rows; ++row) {
	    double x = cdp->cd_xzero + cdp->cd_xincr * (row - cdp->cd_pt_off);

	    (void)printf("%14.7e", x);
	    for (column = 0; column < cdp->cd_columns; ++column) {
		int idx = cdp->cd_columns * row + column;

		switch (cdp->cd_format) {
		case CH_I8:
		    (void)printf(" %3d", cdp->cd_data.s8[idx]);
		    break;

		case CH_X32:
		    (void)printf(" 0x%08" PRIx32, cdp->cd_data.u32[idx]);
		    break;

		case CH_X64:
		    (void)printf(" 0x%013" PRIx64, cdp->cd_data.u64[idx]);
		    break;

		case CH_LF:
		    (void)printf(" %14.7e", cdp->cd_data.lf[idx]);
		    break;
		}
	    }
	    (void)printf("\n");
	}
	free_waveform(cdp);
	(void)printf("\n");
	(void)printf("\n");
    }

    exit(0);
    /*NOTREACHED*/
}
